'Utility to output the end points of an USB device
'just a reduced version of usbview to make it easier for the user
'
$stack 3072
defint a-z

color 15,1 : cls
locate 12,20 : print "ShowEndPoints for DOSUSB"
locate 18,20 : print "(Addresses are starting with 1)"
locate 14,20 : input "Please enter device address: ", devadd%
cls
locate 14,20 : print "Retrieving device descriptor..."

'define structure of URB
type urbtype
  transaction_token as byte 'control (2Dh), in (69h), out (E1h) as hex code
  chain_end_flag  as byte
  dev_add         as byte
  end_point       as byte
  error_code      as byte
  status          as byte  'returned by dosusb
  transaction_flags as word 'reserved
  buffer_off      as word  'for in/out
  buffer_seg      as word  'for in/out
  buffer_length   as word  'for in/out
  actual_length   as word  'for in/out
  setup_buffer_off as word 'for control
  setup_buffer_seg as word 'for control
  start_frame     as word  'reserved
  nr_of_packets   as word  'reserved - iso
  int_interval    as byte  'reserved - int
  error_count     as byte  'reserved
  timeout         as word  'reserved
  next_urb_off    as word  'reserved
  next_urb_seg    as word  'reserved
end type '32 byte long

dim urb as urbtype

type setuptype
   bmRequestType  as byte
   bRequest	  as byte
   wValue	  as word
   wIndex         as word
   wLength        as word
end type

dim device_request as setuptype

type device_descriptor_type
   bLength		as byte
   bDescriptorType	as byte
   bcdUSB		as word
   bDeviceClass		as byte
   bDeviceSubClass	as byte
   bDeviceProtocol	as byte
   bMaxPacketSize	as byte
   idVendor		as word
   idProduct		as word
   bcdDevice		as word
   iManufacturer	as byte
   iProduct		as byte
   iSerialNumber	as byte
   bNumConfigurations	as byte
end type

dim device_descriptor_ptr as device_descriptor_type ptr

type configuration_descriptor_type
   bLength		as byte
   bDescriptorType	as byte
   wTotalLength		as word
   bNumInterfaces	as byte
   bConfigurationValue	as byte
   iConfiguration	as byte
   bmAttributes		as byte
   bMaxPower		as byte
end type

dim configuration_descriptor_ptr as configuration_descriptor_type ptr

type interface_descriptor_type
   bLength		as byte
   bDescriptorType	as byte
   bInterfaceNumber     as byte
   bAlternateSetting    as byte
   bNumEndpoints        as byte
   bInterfaceClass      as byte
   bInterfaceSubClass   as byte
   bInterfaceProtocol   as byte
   iInterface           as byte
end type

dim interface_descriptor_ptr as interface_descriptor_type ptr

type endpoint_descriptor_type
   bLength		as byte
   bDescriptorType	as byte
   bEndpointAddress     as byte
   bmAttributes         as byte
   wMaxPacketSize       as word
   bInterval            as byte
'endpoint descriptor
   cd_bLength           as byte
   cd_bDescriptorType   as byte
   cd_bMaxBurst         as byte
   cd_bmAttributes      as byte
   cd_wBytesPerInterval as word
end type

dim endpoint_descriptor_ptr as endpoint_descriptor_type ptr

dim inbuffer as string*2048
dim inbuffer2 as string*2048

inbuffer=repeat$(1024,chr$(0)) 'return data here
inbuffer2=repeat$(1024,chr$(0)) 'return data here as well

'do 8 byte device_descriptor to determine maxlen of packet
'device descriptor
device_request.bmRequestType=&H80
device_request.bRequest=6
device_request.wValue=&H0100
device_request.wIndex=0
device_request.wLength=&H08

'set up device descriptor request
  urb.transaction_token=&H2D
  urb.chain_end_flag=0
  urb.dev_add=devadd%
  urb.end_point=0
  urb.error_code=0
  urb.status=0
  urb.transaction_flags=0
  urb.buffer_off=varptr(inbuffer)
  urb.buffer_seg=varseg(inbuffer)
  urb.buffer_length=8
  urb.actual_length=8
  urb.setup_buffer_off=varptr(device_request)
  urb.setup_buffer_seg=varseg(device_request)
  urb.start_frame=0
  urb.nr_of_packets=0
  urb.int_interval=0
  urb.error_count=0
  urb.timeout=0
  urb.next_urb_off=0
  urb.next_urb_seg=0

reg 8,varseg(urb)
reg 4,varptr(urb)
call interrupt &H65

'invalid device address?
if urb.error_code=1 then
        cls : locate 14,23
	print "Invalid device address"
        locate 24,23
        input "Press enter to continue", a$
        goto do_end
end if

'transaction error?
if urb.status >1 then
        cls : locate 14,23
	print "Device does not respond"
        locate 24,23
        input "Press enter to continue", a$
        goto do_end
end if


'now determine maxlen
device_descriptor_ptr = varptr32(inbuffer)
maxlen%=@device_descriptor_ptr.bMaxPacketSize
if maxlen%=9 then maxlen%=512 'new for xhci

'full device descriptor
device_request.bmRequestType=&H80
device_request.bRequest=6
device_request.wValue=&H0100
device_request.wIndex=0
device_request.wLength=&H12 '18 'hex 12

'set up device descriptor request
  urb.transaction_token=&H2D
  urb.chain_end_flag=0
  urb.dev_add=devadd%
  urb.end_point=0
  urb.error_code=0
  urb.status=0
  urb.transaction_flags=0
  urb.buffer_seg=varseg(inbuffer)
  urb.buffer_off=varptr(inbuffer)
  urb.buffer_length=18
  urb.actual_length=maxlen%
  urb.setup_buffer_seg=varseg(device_request)
  urb.setup_buffer_off=varptr(device_request)
  urb.start_frame=0
  urb.nr_of_packets=0
  urb.int_interval=0
  urb.error_count=0
  urb.timeout=0
  urb.next_urb_seg=0
  urb.next_urb_off=0

'now call DOSUSB
reg 8,varseg(urb)
reg 4,varptr(urb)
call interrupt &H65

device_descriptor_ptr = varptr32(inbuffer)

'determine length of configuration descriptor
'configuration descriptor
device_request.bmRequestType=&H80
device_request.bRequest=6
device_request.wValue=&H0200
device_request.wIndex=0
device_request.wLength=&H08

'set up configuration descriptor request
  urb.transaction_token=&H2D
  urb.chain_end_flag=0
  urb.dev_add=devadd%
  urb.end_point=0
  urb.error_code=0
  urb.status=0
  urb.transaction_flags=0
  urb.buffer_seg=varseg(inbuffer)
  urb.buffer_off=varptr(inbuffer)
  urb.buffer_length=8 'for configuration descriptor
  urb.actual_length=maxlen%
  urb.setup_buffer_seg=varseg(device_request)
  urb.setup_buffer_off=varptr(device_request)
  urb.start_frame=0
  urb.nr_of_packets=0
  urb.int_interval=0
  urb.error_count=0
  urb.timeout=0
  urb.next_urb_seg=0
  urb.next_urb_off=0

'now call DOSUSB
reg 8,varseg(urb)
reg 4,varptr(urb)
call interrupt &H65

configuration_descriptor_ptr = varptr32(inbuffer)

conf_total_length = @configuration_descriptor_ptr.wTotalLength

'full configuration descriptor
device_request.bmRequestType=&H80
device_request.bRequest=6
device_request.wValue=&H0200
device_request.wIndex=0
device_request.wLength=conf_total_length

'set up configuration descriptor request
  urb.transaction_token=&H2D
  urb.chain_end_flag=0
  urb.dev_add=devadd%
  urb.end_point=0
  urb.error_code=0
  urb.status=0
  urb.transaction_flags=0
  urb.buffer_seg=varseg(inbuffer)
  urb.buffer_off=varptr(inbuffer)
  urb.buffer_length=conf_total_length 'for configuration descriptor
  urb.actual_length=maxlen%
  urb.setup_buffer_seg=varseg(device_request)
  urb.setup_buffer_off=varptr(device_request)
  urb.start_frame=0
  urb.nr_of_packets=0
  urb.int_interval=0
  urb.error_count=0
  urb.timeout=0
  urb.next_urb_seg=0
  urb.next_urb_off=0

'now call DOSUSB
reg 8,varseg(urb)
reg 4,varptr(urb)
call interrupt &H65

configuration_descriptor_ptr = varptr32(inbuffer)


interface_descriptor_ptr = configuration_descriptor_ptr + @configuration_descriptor_ptr.bLength
j=1 'interface counter

while @interface_descriptor_ptr.bDescriptorType = 4

cls
if @interface_descriptor_ptr.iInterface > 0 then
device_request.bmRequestType=&H80
device_request.bRequest=&H06
device_request.wValue=&H0300 + @interface_descriptor_ptr.iInterface
device_request.wIndex=&H0409
device_request.wLength=&H1FF

'set up request
  urb.transaction_token=&H2D
  urb.chain_end_flag=0
  urb.dev_add=devadd%
  urb.end_point=0
  urb.error_code=0
  urb.status=0
  urb.transaction_flags=0
  urb.buffer_seg=varseg(inbuffer2)
  urb.buffer_off=varptr(inbuffer2)
  urb.buffer_length=&H1FF
  urb.actual_length=maxlen%
  urb.setup_buffer_seg=varseg(device_request)
  urb.setup_buffer_off=varptr(device_request)
  urb.start_frame=0
  urb.nr_of_packets=0
  urb.int_interval=0
  urb.error_count=0
  urb.timeout=0
  urb.next_urb_seg=0
  urb.next_urb_off=0

reg 8,varseg(urb)
reg 4,varptr(urb)
call interrupt &H65

end if

endpoint_descriptor_ptr = interface_descriptor_ptr + @interface_descriptor_ptr.bLength

if @endpoint_descriptor_ptr.bDescriptorType = &H21 then
'jump over this descriptor
endpoint_descriptor_ptr = endpoint_descriptor_ptr + @endpoint_descriptor_ptr.bLength
end if

cls

for i=1 to @interface_descriptor_ptr.bNumEndpoints

print : print
a% = @endpoint_descriptor_ptr.bEndpointAddress
a%=a% and &H007F
print "          Endpoint address:" a%;
if @endpoint_descriptor_ptr.bEndpointAddress > 127 then
	print "IN"
else
	print "OUT"
end if
'print "Attributes:      " @endpoint_descriptor_ptr.bmAttributes;
a%=@endpoint_descriptor_ptr.bmAttributes
a%=a% and &H0003
if a%=0 then
print "          Control"
elseif a%=1 then
print "          Isochronous"
elseif a%=2 then
print "          Bulk"
elseif a%=3 then
print "          Interrupt"
end if

endpoint_descriptor_ptr = endpoint_descriptor_ptr + @endpoint_descriptor_ptr.bLength
if maxlen%=512 then endpoint_descriptor_ptr = endpoint_descriptor_ptr + 6 'companion descriptor length

next i 'endpoint loop

print : print
input "          Press enter to continue", a$
cls
interface_descriptor_ptr = endpoint_descriptor_ptr 'advanced above
incr j 'interface counter
wend   'interface loop

do_end:
color 7,0
cls

end

